/*
 *  Copyright (C) 2000 Denis Lemire <denis@lemire.com>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#ifdef HAVE_CONFIG_H
  #include "config.h"
#endif

#ifdef USING_LINUX
  #include <getopt.h>
#endif

#ifndef USING_LINUX
  #include "mygetopt.h"
#endif

#include <stdio.h>
#include "random.h"

struct passwdOptions password;
int quietOutput = 0;

void instructUser(void) {
  printf("passwdgen: You must specify one or more of the `-Aa1@\' options\n");
  printf("Try `passwdgen --help\' for more information.\n");
}

void displayHelp(void) {
  printf("`passwdgen\' generates random passwords based on your criteria.\n");
  printf("\nUsage: passwdgen [OPTIONS]...\n\n");
  printf("password options:\n");
  printf("  -a, --lowercase\tinclude lowercase alphabetical characters\n");
  printf("  -A, --uppercase\tinclude uppercase alphabetical characters\n");
  printf("  -1, --numeric\t\tinclude numerical characters\n");
  printf("  -@, --printable\tinclude nonalphanumeric printable characters (insane)\n");
  printf("\n");
  printf("  -l, --lefthand\tinclude only characters typed with your left hand\n");
  printf("  -r, --righthand\tinclude only characters typed with your right hand\n");
  printf("      --alternate\talternate between left-hand and right-hand characters\n");
  printf("\n");
  printf("      --length=LENGTH\tdesired password length (50 max, 8 default)\n\n");
  printf("other options:\n");
  printf("  -q, --quiet\t\toutput password only, no verbosity\n");
  printf("      --version\t\tdisplay version information and exit\n");
  printf("      --help\t\tdisplay this help and exit\n\n");
  printf("Report bugs to <denis@lemire.com>.\n");
}

void displayVersion(void) {
  printf("%s %s\n", PACKAGE, VERSION);
}

int parseCmdLine(int argc, char *argv[]) {
  int option, option_index;

  static struct option const long_options[] = {
    {"lowercase", no_argument, 0, 'a'},
    {"uppercase", no_argument, 0, 'A'},
    {"numeric", no_argument, 0, '1'},
    {"printable", no_argument, 0, '@'},
    {"lefthand", no_argument, 0, 'l'},
    {"righthand", no_argument, 0, 'r'},
    {"alternate", no_argument, 0, 'b'},
    {"length", required_argument, 0, 'L'},
    {"version", no_argument, 0, 'v'},
    {"help", no_argument, 0, 'h'},
    {"quiet", no_argument, 0, 'q'},
    {NULL, 0, NULL, 0}
  };

  if (argc == 1) {
    instructUser();
    return (1);
  }

  do {
    option_index = 0;
    option = getopt_long (argc, argv, "aA1@lrbq", long_options, &option_index);

    switch (option) {
      case 'a': password.type = password.type + 1; break;
      case 'A': password.type = password.type + 2; break;
      case '1': password.type = password.type + 4; break;
      case '@': password.type = password.type + 8; break;
      case 'L': sscanf(optarg, "%d", &password.length);
        if (password.length > 50) {
          printf("passwdgen: password length must not exceed 50 characters.\n");
          exit();
        } 
      break;
      case 'l': password.hand = 1; break;
      case 'r': password.hand = 2; break;
      case 'b': password.hand = 3; break;
      case 'q': quietOutput = 1; break;
      case 'v': displayVersion(); exit(); break;
      case 'h': displayHelp(); exit(); break;
      case '?': instructUser(); exit(); break;
    }
  } while (option != -1);

  if (password.type == 0) {
    instructUser();
    return (1);
  } else {
    return (0);
  }
}

int main(int argc, char *argv[]) {
  char randPasswd[51];
  int seedStatus;

  password.type = 0;
  password.length = 8;
  password.hand = 0;

  switch (parseCmdLine(argc, argv)) {
    case 0:
      seedStatus = buildPasswd(password.type, password.length,
                               password.hand, &randPasswd[0]);
      break;

    default: return (1); break;
  }

  switch (quietOutput) {
    case 0:
      if (seedStatus == 1)
        fprintf(stderr, "Warning: unable to open /dev/random, using insecure seed.\n");
      printf("Your password is: %s\n", randPasswd);
      break;

    case 1: printf("%s", randPasswd); break;
  }

  return (0);
}
